# flake8: noqa
from configly.config import Config
from configly.interpolators import EnvVarInterpolator, FileInterpolator, Interpolator
from configly.loaders import JsonLoader, TomlLoader, YamlLoader
from configly.registry import register_interpolator, Registry

register_interpolator("ENV", EnvVarInterpolator)
register_interpolator("FILE", FileInterpolator)

__all__ = [
    "Config",
    "EnvVarInterpolator",
    "FileInterpolator",
    "Interpolator",
    "JsonLoader",
    "Registry",
    "Registry",
    "TomlLoader",
    "YamlLoader",
    "register_interpolator",
]



import json

from configly.interpolators import Interpolator
from configly.registry import register_interpolator


class VaultInterpolator(Interpolator):
    def __init__(self, **kwargs):
        try:
            import hvac
        except ImportError:
            raise ImportError("Try installing `configly[vault]` to use the vault interpolator")

        self.client = hvac.Client(**kwargs)

    def __getitem__(self, name):
        import hvac

        try:
            result = self.client.secrets.kv.v1.read_secret(name)
        except hvac.exceptions.InvalidPath:
            raise KeyError(name)
        return json.dumps(result["data"])


register_interpolator("VAULT", VaultInterpolator)



from types import MappingProxyType


class Registry:
    """A registry to allow for non-bundled interpolators and config loaders to be added.

    By default `Config` uses a global registry, to which you can `register_interpolator`.

    If you need more flexibility, you can pass `registry` to any of the `from_*` classmethods
    to use your own registry.

    >>> from configly import Config
    >>> local_registry = Registry()
    >>> config = Config.from_yaml('readthedocs.yml', registry=local_registry)
    """

    def __init__(self):
        self._interpolators = {}

    @property
    def interpolators(self):
        return MappingProxyType(self._interpolators)

    def clear_interpolators(self):
        self._interpolators.clear()

    def register_interpolator(self, name, interpolator_cls, overwrite=False):
        """Register a new interpolator for loading configuration from different sources.

        By default `Config` classes read from a global registry of interpolators. This function
        registers new interpolators to that global registry.

        For example, internally configly registers environment interpolation through a call like:

        >>> from configly import EnvVarInterpolator
        >>> register_interpolator("ENV", EnvVarInterpolator, overwrite=True)
        """
        if name in self._interpolators and not overwrite:
            raise ValueError("Name '{}' is already a registered interpolator".format(name))

        if isinstance(interpolator_cls, type):
            instance = interpolator_cls()
        else:
            instance = interpolator_cls

        self._interpolators[name] = instance


registry = Registry()
register_interpolator = registry.register_interpolator



import json


def quote_string(value: str):
    r"""Add syntactically-correct quotation characters to a string value.

    >>> quote_string("foo")
    '"foo"'

    >>> quote_string("'bar'")
    "'bar'"

    >>> quote_string("ba'z")
    '"ba\'z"'

    >>> quote_string('''q"u'x''')
    '"q\\"u\'x"'

    >>> quote_string('{"log_level": "INFO"}')
    '{"log_level": "INFO"}'
    """
    if (value.startswith('"') and value.endswith('"')) or (
        value.startswith("'") and value.endswith("'")
    ):
        # NOTE:
        #   Single quotes aren't always accepted as valid syntax.
        #   And just because a value is surrounded by quotes, does not mean it is correctly quoted.
        #   In that case we would rely on the user of this fn to raise a helpful error.
        return value

    try:
        json.loads(value)
    except json.decoder.JSONDecodeError:
        # If the value is unable to load via json.loads, this means we have to quote it
        # and perform any necessary escaping.
        return json.dumps(value)

    return value



import functools
import os

from configly.interpolators import Interpolator


class FileInterpolator(Interpolator):
    yaml_safe = False

    @functools.lru_cache()
    def __getitem__(self, name):
        if not os.path.exists(name):
            raise KeyError("{} file does not exist.")

        with open(name, mode="rb") as f:
            return f.read().decode("utf-8")



import os

from configly.interpolators import Interpolator


class EnvVarInterpolator(Interpolator):
    def __getitem__(self, name):
        return os.environ[name]

    def get(self, name, default=None):
        return os.environ.get(name, default)



import importlib
import json
from typing import TypeVar


class YamlLoader:
    def __init__(self):
        try:
            from ruamel.yaml import YAML
        except ImportError:
            raise ImportError("Install `configly[yaml]` to use the yaml loader.")

        yaml = YAML(typ="safe")

        self.decoder = yaml

    def load(self, value):
        return self.decoder.load(value)

    def loads(self, value):
        return self.load(value)

    def load_value(self, value):
        return self.decoder.load(value)


class JsonLoader:
    def __init__(self):
        self.decoder = json.JSONDecoder()

    def load(self, value):
        return json.load(value)

    def loads(self, value: str):
        return json.loads(value)

    def load_value(self, value):
        try:
            return self.decoder.decode(value)
        except ValueError:
            return value


def _tomli_parse_value(loader, value):
    return loader._parser.parse_value(value, 0, float)[1]


def _toml_parse_value(loader, value):
    decoder = loader.TomlDecoder()
    return decoder.load_value(value)[0]


class TomlLoader:
    _packages = [
        ("tomllib", _tomli_parse_value),
        ("tomli", _tomli_parse_value),
        ("toml", _toml_parse_value),
    ]

    def __init__(self):
        loader = None
        decoder = None
        for package, package_decoder in self._packages:
            try:
                loader = importlib.import_module(package)
            except ImportError:
                continue
            else:
                decoder = package_decoder
                break

        if loader is None:
            raise ImportError(
                "No toml parser available. One is included in python 3.11's stdlib default, or "
                "alternatively install `configly[tomli]` or configly[toml]`."
            )

        self.loader = loader
        self.decoder = decoder

    def load(self, value):
        return self.loader.load(value)

    def loads(self, value):
        return self.loader.loads(value)

    def load_value(self, value):
        try:
            return self.decoder(self.loader, value)
        except ValueError:
            return value


Loader = TypeVar("Loader", YamlLoader, JsonLoader, TomlLoader)



import re
from collections.abc import Iterable, Mapping

from configly.registry import registry
from configly.utilities import quote_string

INTERPOLATION_REGEX = re.compile(r"(.*)\<%\s*(\w+)\[([\w.]+)(?:,\s*(.+))?\]\s*%>(.*)")


def post_process(loader, value, registry=registry):
    if isinstance(value, Mapping):
        result = {}
        for key, item in value.items():
            result[key] = post_process(loader, item, registry=registry)
        return result

    elif isinstance(value, Iterable) and not isinstance(value, str):
        result = []
        for item in value:
            result.append(post_process(loader, item, registry=registry))
        return result

    else:
        # Repeatedly evaluate the string until there are no interpolation blocks
        while True:
            # The post-interpolation of previous values might coerce them into
            # concrete values, on which no further processing is necessary.
            if not isinstance(value, str):
                break

            match = re.match(INTERPOLATION_REGEX, value)
            if not match:
                break

            groups = match.groups()
            pre, interpolation_type, var_name, default, post = groups

            if interpolation_type not in registry.interpolators:
                raise ValueError("Unrecognized interpolator type: {}".format(interpolation_type))

            interpolator = registry.interpolators[interpolation_type]
            if default is None:
                try:
                    var_value = interpolator[var_name]
                except KeyError:
                    raise ValueError(
                        "The requested {} value '{}' was not found".format(
                            interpolation_type.lower(), var_name
                        )
                    )
            else:
                var_value = interpolator.get(var_name, default)

            result = pre + var_value + post
            if len(result) and not result[0].isalnum():
                # If the first character of `pre` is not alphanumeric, safely quote the result.
                result = quote_string(result)

            if getattr(interpolator, "yaml_safe", True):
                value = loader.load_value(result)
            else:
                value = result

        return value



import abc


class Interpolator(abc.ABC):
    """ABC to define the interface required by an interpolator.

    It is not required to subclass `Interpolator`, but it *does* provide the interface
    and ensures the class implements it.
    """

    @abc.abstractmethod
    def __getitem__(self, name):
        """Override this method to implement a method to get the value for a piece of config.

        This method should return a `KeyError` when the value cannot be found.
        """

    def get(self, name, default=None):
        """Implement get operation with a default.

        Override this method to get more tailored behavior.
        """
        try:
            return self[name]
        except KeyError:
            return default


# isort: split
from configly.interpolators.docker_secret import DockerSecretInterpolator  # noqa
from configly.interpolators.env import EnvVarInterpolator  # noqa
from configly.interpolators.file import FileInterpolator  # noqa

__all__ = ["EnvVarInterpolator", "FileInterpolator", "DockerSecretInterpolator", "Interpolator"]



import copy
from collections.abc import Mapping
from typing import Optional

from configly.loaders import JsonLoader, Loader, TomlLoader, YamlLoader
from configly.process import post_process
from configly.registry import registry


class Config:
    """Container for configuration.

    >>> config = Config({"a": 1, "b": {"c": 2}})
    >>> config.a
    1
    >>> config.b.c
    2
    """

    def __init__(self, value=None, _src_input=None, _loader=None, _registry=registry):
        if value is None:
            value = {}
        self._value = value

        self._src_input = _src_input
        self._loader = _loader
        self._registry = _registry

    @classmethod
    def from_loader(
        cls,
        loader: Loader,
        file: Optional[str] = None,
        content: Optional[str] = None,
        registry=registry,
    ):
        result = {}

        if file:
            with open(file, "rb") as f:
                result = loader.load(f)

        if content:
            result = loader.loads(content)

        output = post_process(loader=loader, value=result, registry=registry)
        return cls(output, _src_input=result, _loader=loader, _registry=registry)

    @classmethod
    def from_yaml(
        cls, file: Optional[str] = None, *, content: Optional[str] = None, registry=registry
    ):
        """Open a yaml `file` and load it into the resulting config object."""
        return cls.from_loader(YamlLoader(), file=file, content=content, registry=registry)

    @classmethod
    def from_json(
        cls, file: Optional[str] = None, content: Optional[str] = None, registry=registry
    ):
        """Open a toml `file` and load it into the resulting config object."""
        return cls.from_loader(JsonLoader(), file=file, content=content, registry=registry)

    @classmethod
    def from_toml(
        cls, file: Optional[str] = None, content: Optional[str] = None, registry=registry
    ):
        """Open a toml `file` and load it into the resulting config object."""
        return cls.from_loader(TomlLoader(), file=file, content=content, registry=registry)

    def refresh(self):
        """Reevaluate the interpolation of variable values in the given sub-config.

        This will be particularly useful for values which are coming from sources
        where the value might change.
        """
        update = post_process(loader=self._loader, value=self._src_input, registry=self._registry)
        self._value.update(update)

    def to_dict(self):
        """Return a `dict` equivalent of the config object.

        Roughly equivalent to
        >>> dict(Config({1:1})) == Config({1:1}).to_dict()
        True
        """
        return copy.deepcopy(self._value)

    def __iter__(self):
        for key, value in self._value.items():
            if isinstance(value, Mapping):
                value = self.__class__(
                    value,
                    _src_input=self._src_input and self._src_input[key],
                    _loader=self._loader,
                    _registry=self._registry,
                )

            yield key, value

    def __getitem__(self, attr):
        try:
            value = self._value[attr]
        except KeyError:
            raise KeyError("'{}' not found in: {}.".format(attr, self))

        if isinstance(value, Mapping):
            return self.__class__(
                value,
                _src_input=self._src_input and self._src_input[attr],
                _loader=self._loader,
                _registry=self._registry,
            )
        return value

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError as e:
            if hasattr(self._value, name):
                return getattr(self._value, name)
            raise AttributeError(str(e))

    def __eq__(self, other):
        if isinstance(other, type(self)):
            return self._value == other._value
        return self._value == other

    def __repr__(self):
        return "{0.__class__.__name__}({0._value})".format(self)



import os

from configly.interpolators import Interpolator
from configly.registry import registry


class DockerSecretInterpolator(Interpolator):
    def __getitem__(self, name):
        env_file = f"{name.upper()}_FILE"

        file_value = os.environ.get(env_file)
        if file_value is not None:
            # When the _FILE env var is set, the actual value is found by getting
            # the contents of the file at the location indicated by the env var.
            with open(file_value, "rb") as f:
                raw_value = f.read()

            value = raw_value.decode("utf-8")
            return value

        return os.environ[name]

    @classmethod
    def register(cls, registry=registry, overwrite=False):
        registry.register_interpolator("DOCKER_SECRET", cls, overwrite=overwrite)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['configly', 'configly.interpolators']

package_data = \
{'': ['*']}

extras_require = \
{'toml': ['toml'],
 'tomli': ['tomli'],
 'vault': ['hvac'],
 'yaml': ['ruamel.yaml']}

setup_kwargs = {
    'name': 'configly',
    'version': '1.0.0',
    'description': '',
    'long_description': '![Github Actions](https://github.com/schireson/configly/actions/workflows/build.yml/badge.svg) [![codecov](https://codecov.io/gh/schireson/configly/branch/main/graph/badge.svg)](https://codecov.io/gh/schireson/configly) [![Documentation Status](https://readthedocs.org/projects/configly/badge/?version=latest)](https://configly.readthedocs.io/en/latest/?badge=latest)\n\n## TL;DR\n\n```yaml\n# config.yml\nfoo:\n  bar: <% ENV[REQUIRED] %>\n  baz: <% ENV[OPTIONAL, true] %>\nlist_of_stuff:\n  - fun<% ENV[NICE, dament] %>al\n  - fun<% ENV[AGH, er] %>al\n  - more/<% ENV[THAN, er] %>/one/<% ENV[interpolation, er] %>!\n```\n\n```python\n# app.py\nconfig = Config.from_yaml(\'config.yml\')\n\nprint(config.foo.bar)\nprint(config.foo[\'baz\'])\nfor item in config.list_of_stuff:\n    print(item)\n```\n\n```bash\npip install configly[yaml]\n```\n\n## Introduction\n\nLoading configuration is done in every (application) project, and yet it is often\noverlooked and condidered too easy or straightforward to bother using a library\nto manage doing it.\n\nTherefore, we often see code like this:\n\n```python\n# config.py\nimport os\n\n# Maybe it\'s following 12factor and loading all the config from the environment.\nconfig = {\n    \'log_level\': os.getenv(\'LOG_LEVEL\'),\n    \'database\': {\n        # At least here, I can nest values if I want to organize things.\n        \'password\': os.environ[\'DATABASE_PASSWORD\'],\n        \'port\': int(os.environ[\'DATABASE_PORT\']),\n    }\n}\n```\n\nor this\n\n```python\n# config.py\nimport os\n\nclass Config:\n    log_level = os.getenv(\'LOG_LEVEL\')\n\n    # Here it\'s not so easy to namespace\n    database_password = os.environ[\'DATABASE_PASSWORD\']\n    database_port = int(os.environ[\'DATABASE_PORT\'])\n\n\n# Oh goodness!\nclass DevConfig(Config):\n    environment = \'dev\'\n```\n\nor this\n\n```python\nimport configparser\n# ...🤢... Okay I dont even want to get into this one.\n```\n\nAnd this is all assuming that everyone is loading configuration at the outermost entrypoint!\nThe two worst possible outcomes in configuration are:\n\n- You are loading configuration lazily and/or deeply within your application, such that it\n  hits a critical failure after having seemingly successfully started up.\n- There is not a singular location at which you can go to see all configuration your app might\n  possibly be reading from.\n\n## The pitch\n\n`Configly` asserts configuration should:\n\n- Be centralized\n  - One should be able to look at one file to see all (env vars, files, etc) which must exist for the\n    application to function.\n- Be comprehensive\n  - One should not find configuration being loaded secretly elsewhere\n- Be declarative/static\n  - code-execution (e.g. the class above) in the definition of the config inevitably makes it\n    hard to interpret, as the config becomes more complex.\n- Be namespacable\n  - One should not have to prepend `foo_` namespaces to all `foo` related config names\n- Be loaded, once, at app startup\n  - (At least the _definition_ of the configuration you\'re loading)\n- (Ideally) have structured output\n  - If something is an `int`, ideally it would be read as an int.\n\nTo that end, the `configly.Config` class exposes a series of classmethods from which your config\ncan be loaded. It\'s largely unimportant what the input format is, but we started with formats\nthat deserialize into at least `str`, `float`, `int`, `bool` and `None` types.\n\n```python\n# Currently supported input formats.\nconfig = Config.from_yaml(\'config.yml\')\nconfig = Config.from_json(\'config.json\')\nconfig = Config.from_toml(\'config.toml\')\n```\n\nGiven an input `config.yml` file:\n\n```yaml\n# config.yml\nfoo:\n  bar: <% ENV[REQUIRED] %>\n  baz: <% ENV[OPTIONAL, true] %>\nlist_of_stuff:\n  - fun<% ENV[NICE, dament] %>al\n  - fun<% ENV[AGH, er] %>al\n  - more/<% ENV[THAN, er] %>/one/<% ENV[interpolation, er] %>!\n```\n\nA number of things are exemplified in the example above:\n\n- Each `<% ... %>` section indicates an interpolated value, the interpolation can\n  be a fragment of the overall value, and multiple values can be interpolated\n  within a single value.\n\n- `ENV` is an "interpolator" which knows how to obtain environment variables\n\n- `[VAR]` Will raise an error if that piece of config is not found, whereas\n  `[VAR, true]` will default `VAR` to the value after the comma\n\n- Whatever the final value is, it\'s interpreted as a literal value in the\n  format of the file which loads it. I.E. `true` -> python `True`, `1` ->\n  python `1`, and `null` -> python `None`.\n\nNow that you\'ve loaded the above configuration:\n\n```python\n# app.py\nconfig = Config.from_yaml(\'config.yml\')\n\n# You can access namespaced config using dot access\nprint(config.foo.bar)\n\n# You have use index syntax for dynamic, or non-attribute-safe key values.\nprint(config.foo[\'baz\'])\n\n# You can iterate over lists\nfor item in config.list_of_stuff:\n    print(item)\n\n# You can *generally* treat key-value maps as dicts\nfor key, value in config.foo.items():\n    print(key, value)\n\n# You can *actually* turn key-value maps into dicts\ndict(config.foo) == config.foo.to_dict()\n```\n\n## Installing\n\n```bash\n# Basic installation\npip install configly\n\n# To use the yaml config loader\npip install configly[yaml]\n\n# To use the toml config loader\npip install configly[toml]\n\n# To use the vault config loader\npip install configly[vault]\n```\n',
    'author': 'Dan Cardin',
    'author_email': 'ddcardin@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/schireson/configly',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'extras_require': extras_require,
    'python_requires': '>=3.6.2,<4',
}


setup(**setup_kwargs)
